const builtin = @import("builtin");

const ona = @import("./ona.zig");

const std = @import("std");

pub const Error = error {
	UnavailableResource,
};

pub fn Functor(comptime Output: type, comptime input_types: []const type) type {
	const InputTuple = std.meta.Tuple(input_types);

	return struct {
		context: *const anyopaque,
		apply_with_context: *const fn (*const anyopaque, InputTuple) Output,

		const Self = @This();

		pub fn apply(self: *const Self, inputs: InputTuple) Output {
			return self.apply_with_context(self.context, inputs);
		}

		pub fn bind(comptime State: type, state: *const State, comptime invoke: anytype) Self {
			const is_zero_aligned = @alignOf(State) == 0;

			return .{
				.context = if (is_zero_aligned) state else @ptrCast(state),

				.apply_with_context = struct {
					fn invoke_concrete(context: *const anyopaque, inputs: InputTuple) Output {
						if (is_zero_aligned) {
							return @call(.auto, invoke, .{@as(*const State, @ptrCast(context))} ++ inputs);
						}

						return switch (@typeInfo(@typeInfo(@TypeOf(invoke)).Fn.return_type.?)) {
							.ErrorUnion => try @call(.auto, invoke, .{@as(*const State, @ptrCast(@alignCast(context)))} ++ inputs),
							else => @call(.auto, invoke, .{@as(*const State, @ptrCast(@alignCast(context)))} ++ inputs),
						};
					}
				}.invoke_concrete,
			};
		}

		pub fn bind_fn(comptime invoke: anytype) Self {
			return .{
				.context = undefined,

				.apply_with_context = struct {
					fn invoke_concrete(_: *const anyopaque, inputs: InputTuple) Output {
						return @call(.auto, invoke, inputs);
					}
				}.invoke_concrete,
			};
		}
	};
}

pub fn Generator(comptime Output: type, comptime input_types: []const type) type {
	const InputTuple = std.meta.Tuple(input_types);

	return struct {
		context: *anyopaque,
		yield_with_context: *const fn (*anyopaque, InputTuple) Output,

		const Self = @This();

		pub fn bind(comptime State: type, state: *State, comptime invoke: anytype) Self {
			const is_zero_aligned = @alignOf(State) == 0;

			return .{
				.context = if (is_zero_aligned) state else @ptrCast(state),

				.yield_with_context = struct {
					fn invoke_concrete(context: *anyopaque, inputs: InputTuple) Output {
						if (is_zero_aligned) {
							return @call(.auto, invoke, .{@as(*State, @ptrCast(context))} ++ inputs);
						}

						return switch (@typeInfo(@typeInfo(@TypeOf(invoke)).Fn.return_type.?)) {
							.ErrorUnion => try @call(.auto, invoke, .{@as(*State, @ptrCast(@alignCast(context)))} ++ inputs),
							else => @call(.auto, invoke, .{@as(*State, @ptrCast(@alignCast(context)))} ++ inputs),
						};
					}
				}.invoke_concrete,
			};
		}

		pub fn bind_fn(comptime invoke: anytype) Self {
			return .{
				.context = undefined,

				.yield_with_context = struct {
					fn invoke_concrete(_: *const anyopaque, inputs: InputTuple) Output {
						return @call(.auto, invoke, inputs);
					}
				}.invoke_concrete,
			};
		}

		pub fn yield(self: *const Self, inputs: InputTuple) Output {
			return self.yield_with_context(self.context, inputs);
		}
	};
}

pub const NullWritable = struct {
	written: usize = 0,

	pub fn write(self: *NullWritable, buffer: []const u8) Error!usize {
		self.written += buffer.len;

		return buffer.len;
	}

	pub fn writer(self: *NullWritable) Writer {
		return Writer.bind(NullWritable, self, write);
	}
};

pub const Reader = Generator(Error!usize, &.{[]u8});

pub const Writer = Generator(Error!usize, &.{[]const u8});

pub fn alloc_read(input: ona.io.Reader, allocator: std.mem.Allocator) []u8 {
	const buffer = ona.stack.Sequential(u8){.allocator = allocator};

	errdefer buffer.deinit();

	const streamed = try stream_all(input.reader(), buffer.writer());

	return buffer.to_allocation(streamed);
}

pub const bits_per_byte = 8;

pub fn skip_n(input: Reader, distance: u64) Error!void {
	var buffer = @as([512]u8, undefined);
	var remaining = distance;

	while (remaining != 0) {
		const read = try input.yield(.{buffer[0 .. @min(remaining, buffer.len)]});

		if (read == 0) {
			return error.UnavailableResource;
		}

		remaining -= read;
	}
}

pub fn stream_all(input: Reader, output: Writer) Error!usize {
	var buffer = @as([512]u8, undefined);
	var copied = @as(usize, 0);

	while (true) {
		const read = try input.apply(.{&buffer});

		if (read == 0) {
			return copied;
		}

		if (try output.apply(.{buffer[0 .. read]}) != read) {
			return error.UnavailableResource;
		}

		copied += read;
	}
}

pub fn stream_n(input: Reader, output: Writer, limit: usize) Error!usize {
	var buffer = @as([512]u8, undefined);
	var remaining = limit;

	while (true) {
		const read = try input.yield(.{buffer[0 .. @min(remaining, buffer.len)]});

		if (read == 0) {
			return limit - remaining;
		}

		if (try output.yield(.{buffer[0 .. read]}) != read) {
			return error.UnavailableResource;
		}

		remaining -= read;
	}
}

pub fn write_all(writer: Writer, utf8: []const u8) Error!void {
	if (try writer.yield(.{utf8}) != utf8.len) {
		return error.UnavailableResource;
	}
}
