pub const ascii = @import("./ascii.zig");

pub const asyncio = @import("./asyncio.zig");

pub const dag = @import("./dag.zig");

pub const hashes = @import("./hashes.zig");

pub const heap = @import("./heap.zig");

pub const files = @import("./files.zig");

pub const io = @import("./io.zig");

pub const map = @import("./map.zig");

pub const scalars = @import("./scalars.zig");

pub const slices = @import("./slices.zig");

pub const stack = @import("./stack.zig");

const std = @import("std");

pub const utf8 = @import("./utf8.zig");

pub const World = @import("./World.zig");

pub const lina = @import("./lina.zig");

pub const App = struct {
	events: *const Events,
	target_frame_time: f64,
	elapsed_time: f64,
	is_running: bool,

	pub const Events = struct {
		load: World.Event,
		pre_update: World.Event,
		update: World.Event,
		post_update: World.Event,
		render: World.Event,
		finish: World.Event,
		exit: World.Event,
	};

	pub fn quit(self: *App) void {
		self.is_running = false;
	}
};

fn Channel(comptime Message: type) type {
	return struct {
		buffers: [2]stack.Sequential(Message),
		swap_index: u1 = 0,
		ticks: u1 = 0,

		const Self = @This();

		fn cleanup(channel: Write(Self)) void {
			channel.state.deinit();
		}

		pub fn deinit(self: *Self) void {
			for (&self.buffers) |*buffer| {
				buffer.deinit();
			}

			self.* = undefined;
		}

		fn swap(channel: Write(Self)) void {
			channel.state.ticks = scalars.add(channel.state.ticks, 1) orelse 0;

			if (channel.state.ticks == 0) {
				channel.state.swap_index ^= 1;

				channel.state.buffers[channel.state.swap_index].clear();
			}
		}

		fn init(allocator: std.mem.Allocator) Self {
			return .{
				.buffers = .{
					.{.allocator = allocator},
					.{.allocator = allocator},
				},
			};
		}

		fn messages(self: Self) []const Message {
			return self.buffers[self.swap_index ^ 1].values;
		}

		fn push(self: *Self, message: Message) std.mem.Allocator.Error!void {
			try self.buffers[self.swap_index].push_grow(message);
		}
	};
}

pub fn Exclusive(comptime values: []const type) type {
	comptime var qualifieds: [values.len]type = undefined;

	for (&qualifieds, values) |*qualified, Value| {
		qualified.* = switch (@typeInfo(Value)) {
			.Optional => @Type(.{
				.Optional = .{
					.child = .{
						.Pointer = .{
							.is_allowzero = false,
							.sentinel = null,
							.address_space = .generic,
							.is_volatile = false,
							.alignment = @alignOf(Value),
							.size = .One,
							.child = Value,
							.is_const = false,
						},
					},
				},
			}),

			else => @Type(.{
				.Pointer = .{
					.is_allowzero = false,
					.sentinel = null,
					.address_space = .generic,
					.is_volatile = false,
					.alignment = @alignOf(Value),
					.size = .One,
					.child = Value,
					.is_const = false,
				},
			}),
		};
	}

	const States = std.meta.Tuple(&qualifieds);

	return struct {
		states: States,
		world: *World,

		const Self = @This();

		pub const Param = struct {
			states: States,
			world: *World,
		};

		pub fn bind(context: World.BindContext) std.mem.Allocator.Error!Param {
			var param = Param{
				.world = context.world,
				.states = undefined,
			};

			inline for (&param.states, values) |*state, Value| {
				const has_state = try context.register_writable_state_access(Value);

				state.* = switch (@typeInfo(Value)) {
					.Optional => has_state,

					else => has_state orelse {
						@panic(std.fmt.comptimePrint("attempt to use exclusive {s} that has not yet been set", .{
							@typeName(Value),
						}));
					},
				};
			}

			return param;
		}

		pub fn init(param: *Param) Self {
			return .{
				.states = param.states,
				.world = param.world,
			};
		}

		pub const thread_restriction = .main;
	};
}

pub const LaunchFlag = enum {
	dump_shader_translations,

	var args = [_]?[]const u8{null} ** std.enums.values(LaunchFlag).len;

	const values = std.enums.values(LaunchFlag);
};

pub const Options = struct {
	tick_rate: u64,
	execution: Execution,
	middleware: []const *const Setup,

	pub const Execution = union (enum) {
		single_threaded,
		thread_share: f32,
	};
};

pub fn Params(comptime Value: type) type {
	if (!@hasDecl(Value, "Param")) {
		@compileError("System parameters must have a Params type declaration");
	}

	return struct {
		has_head: ?*Node = null,
		has_tail: ?*Node = null,

		pub const Node = struct {
			param: Value.Param,
			has_prev: ?*Node = null,
			has_next: ?*Node = null,
		};
	};
}

pub fn Pool(comptime Value: type) type {
	return struct {
		entries: stack.Sequential(Entry),
		first_free_index: usize = 0,

		const Entry = union (enum) {
			free_index: usize,
			occupied: Value,
		};

		pub const Values = struct {
			cursor: usize = 0,
			pool: *const Self,

			pub fn next(self: *Values) ?*Value {
				while (self.cursor < self.pool.entries.len()) {
					defer self.cursor += 1;

					switch (self.pool.entries.values[self.cursor]) {
						.free_index => {
							continue;
						},

						.occupied => |*occupied| {
							return occupied;
						},
					}
				}

				return null;
			}
		};

		const Self = @This();

		pub fn deinit(self: *Self) void {
			self.entries.deinit();

			self.* = undefined;
		}

		pub fn get(self: *Self, key: usize) ?*Value {
			if (key >= self.entries.len()) {
				return null;
			}

			return switch (self.entries.values[key]) {
				.free_index => null,
				.occupied => |*occupied| occupied,
			};
		}

		pub fn init(allocator: std.mem.Allocator) Self {
			return .{
				.entries = .{.allocator = allocator},
			};
		}

		pub fn insert(self: *Self, value: Value) std.mem.Allocator.Error!usize {
			const entries_count = self.entries.len();

			if (self.first_free_index == entries_count) {
				try self.entries.push_grow(.{.occupied = value});

				self.first_free_index += 1;

				return entries_count;
			}

			const insersion_index = self.first_free_index;

			self.first_free_index = self.entries.values[self.first_free_index].free_index;
			self.entries.values[insersion_index] = .{.occupied = value};

			return insersion_index;
		}

		pub fn remove(self: *Self, key: usize) ?Value {
			if (key >= self.entries.len()) {
				return null;
			}

			switch (self.entries.values[key]) {
				.free_index => {
					return null;
				},

				.occupied => |occupied| {
					self.entries.values[key] = .{.free_index = self.first_free_index};
					self.first_free_index = key;

					return occupied;
				},
			}
		}

		pub fn values(self: *const Self) Values {
			return .{
				.pool = self,
			};
		}
	};
}

test "pooling" {
	var pool = Pool(i32).init(std.testing.allocator);

	defer pool.deinit();

	try std.testing.expectEqual(0, pool.first_free_index);
	try std.testing.expectEqual(null, pool.get(0));
	try std.testing.expectEqual(null, pool.remove(0));
	try std.testing.expectEqual(0, try pool.insert(69));
	try std.testing.expectEqual(1, pool.first_free_index);
	try std.testing.expectEqual(1, try pool.insert(420));
	try std.testing.expectEqual(69, pool.get(0).?.*);
	try std.testing.expectEqual(420, pool.get(1).?.*);
	try std.testing.expectEqual(69, pool.remove(0).?);
	try std.testing.expectEqual(0, pool.first_free_index);
	try std.testing.expectEqual(null, pool.get(0));
	try std.testing.expectEqual(null, pool.remove(0));
	try std.testing.expectEqual(0, try pool.insert(36));
	try std.testing.expectEqual(36, pool.get(0).?.*);
}

pub fn Read(comptime Value: type) type {
	return Shared(Value, .{
		.thread_restriction = thread_restriction(Value),
		.read_only = true,
	});
}

pub fn Receive(comptime Message: type) type {
	const TypedChannel = Channel(Message);

	return struct {
		channel: *const TypedChannel,

		const Self = @This();

		pub const Param = struct {
			channel: *const TypedChannel,
		};

		pub fn bind(context: World.BindContext) std.mem.Allocator.Error!Param {
			return .{
				.channel = (try context.register_readable_state_access(TypedChannel)) orelse set: {
					try context.world.set_state(TypedChannel.init(heap.allocator));

					break: set (try context.register_readable_state_access(TypedChannel)).?;
				},
			};
		}

		pub fn init(param: *Param) Self {
			return .{
				.channel = param.channel,
			};
		}

		pub fn messages(self: Self) []const Message {
			return self.channel.messages();
		}
	};
}

pub fn Send(comptime Message: type) type {
	const TypedChannel = Channel(Message);

	return struct {
		channel: *TypedChannel,

		const Self = @This();

		pub const Param = struct {
			channel: *TypedChannel,
		};

		pub fn bind(context: World.BindContext) std.mem.Allocator.Error!Param {
			return .{
				.channel = (try context.register_writable_state_access(TypedChannel)) orelse set: {
					try context.world.set_state(TypedChannel.init(heap.allocator));

					const app = context.world.get_state(App) orelse {
						@panic("Send system parameters depend on a " ++ @typeName(App) ++ " state to work");
					};

					try context.world.on_event(app.events.post_update, system_fn(TypedChannel.swap), .{
						.label = "swap channel of " ++ @typeName(Message),
					});

					try context.world.on_event(app.events.exit, system_fn(TypedChannel.cleanup), .{
						.label = "clean up channel of " ++ @typeName(Message),
					});

					break: set (try context.register_writable_state_access(TypedChannel)).?;
				},
			};
		}

		pub fn init(param: *Param) Self {
			return .{
				.channel = param.channel,
			};
		}

		pub fn push(self: Self, message: Message) std.mem.Allocator.Error!void {
			try self.channel.push(message);
		}
	};
}

pub const Setup = fn (*World, App.Events) anyerror!void;

pub const ShareInfo = struct {
	thread_restriction: ThreadRestriction,
	read_only: bool,
};

pub fn Shared(comptime Value: type, comptime info: ShareInfo) type {
	const value_info = @typeInfo(Value);

	const Qualified = switch (value_info) {
		.Optional => @Type(.{
			.Optional = .{
				.child = .{
					.Pointer = .{
						.is_allowzero = false,
						.sentinel = null,
						.address_space = .generic,
						.is_volatile = false,
						.alignment = @alignOf(Value),
						.size = .One,
						.child = Value,
						.is_const = info.read_only,
					},
				},
			},
		}),

		else => @Type(.{
			.Pointer = .{
				.is_allowzero = false,
				.sentinel = null,
				.address_space = .generic,
				.is_volatile = false,
				.alignment = @alignOf(Value),
				.size = .One,
				.child = Value,
				.is_const = info.read_only,
			},
		}),
	};

	return struct {
		state: Qualified,

		const Self = @This();

		pub const Param = struct {
			state: Qualified,
		};

		pub fn bind(context: World.BindContext) std.mem.Allocator.Error!Param {
			const thread_restriction_name = switch (info.thread_restriction) {
				.main => "main thread-restricted ",
				.none => ""
			};

			const state = switch (info.read_only) {
				true => (try context.register_readable_state_access(Value)),
				false => (try context.register_writable_state_access(Value)),
			};

			return .{
				.state = switch (value_info) {
					.Optional => state,

					else => state orelse {
						@panic(std.fmt.comptimePrint("attempt to use {s}{s} {s} that has not yet been set", .{
							thread_restriction_name,
							if (info.read_only) "read-only" else "read-write",
							@typeName(Value),
						}));
					},
				},
			};
		}

		pub fn init(param: *Param) Self {
			return .{
				.state = param.state,
			};
		}

		pub const thread_restriction = info.thread_restriction;
	};
}

pub const SystemInfo = struct {
	execute: *const fn ([]const *const Parameter, *const [max_parameters]*anyopaque) anyerror!void,
	parameters: [max_parameters]*const Parameter = undefined,
	parameter_count: u4 = 0,
	thread_restriction: ThreadRestriction = .none,

	pub const Parameter = struct {
		thread_restriction: ThreadRestriction,
		init: *const fn (*anyopaque, *anyopaque) void,
		bind: *const fn (std.mem.Allocator, World.BindContext) std.mem.Allocator.Error!*anyopaque,
		unbind: *const fn (std.mem.Allocator, *anyopaque, World.UnbindContext) void,
	};

	pub const max_parameters = 16;

	pub fn used_parameters(self: *const SystemInfo) []const *const Parameter {
		return self.parameters[0 .. self.parameter_count];
	}
};

pub const SystemOrder = struct {
	label: []const u8 = "",
	run_after: []const *const SystemInfo = &.{},
	run_before: []const *const SystemInfo = &.{},
};

pub const ThreadRestriction = enum {
	none,
	main,
};

pub const TypeID = enum (usize) { _ };

pub fn Write(comptime Value: type) type {
	return Shared(Value, .{
		.thread_restriction = thread_restriction(Value),
		.read_only = false,
	});
}

pub fn launch_arg(flag: LaunchFlag) ?[]const u8 {
	return LaunchFlag.args[@intFromEnum(flag)];
}

fn parameter_type(comptime Value: type) *const SystemInfo.Parameter {
	const ValueParams = Params(Value);

	if (@sizeOf(Value) == 0) {
		@compileError("System parameters must have a non-zero size");
	}

	const parameters = struct {
		fn bind(allocator: std.mem.Allocator, context: World.BindContext) std.mem.Allocator.Error!*anyopaque {
			const value_name = @typeName(Value);

			if (!@hasDecl(Value, "bind")) {
				@compileError(
					"a `bind` declaration on " ++
					value_name ++
					" is requied for parameter types with a `Param` declaration");
			}

			const bind_type = @typeInfo(@TypeOf(Value.bind));

			if (bind_type != .Fn) {
				@compileError("`bind` declaration on " ++ value_name ++ " must be a fn");
			}

			if (bind_type.Fn.params.len != 1 or bind_type.Fn.params[0].type.? != World.BindContext) {
				@compileError(
					"`bind` fn on " ++
					value_name ++
					" must accept " ++
					@typeName(World.BindContext) ++
					" as it's one and only argument");
			}

			const params_node = try allocator.create(ValueParams.Node);

			params_node.* = .{
				.param = switch (bind_type.Fn.return_type.?) {
					Value.Param => Value.bind(context),
					std.mem.Allocator.Error!Value.Param => try Value.bind(context),

					else => @compileError(std.fmt.comptimePrint("`bind` fn on {s} must return {s} or {s}", .{
						@typeName(Value),
						@typeName(Value.Param),
						@typeName(std.mem.Allocator.Error!Value.Param)
					})),
				},
			};

			if (context.world.get_state(ValueParams)) |value_params| {
				if (value_params.has_tail) |tail| {
					tail.has_next = params_node;
				}

				params_node.has_prev = value_params.has_tail;
				value_params.has_tail = params_node;
			} else {
				try context.world.set_state(ValueParams{
					.has_head = params_node,
					.has_tail = params_node,
				});
			}

			return @ptrCast(params_node);
		}

		fn init(argument: *anyopaque, erased_node: *anyopaque) void {
			const value_name = @typeName(Value);

			if (!@hasDecl(Value, "init")) {
				@compileError("an `init` declaration on " ++ value_name ++ " is requied for parameter types");
			}

			const init_type = @typeInfo(@TypeOf(Value.init));

			if (init_type != .Fn) {
				@compileError("`init` declaration on " ++ value_name ++ " must be a fn");
			}

			if (init_type.Fn.return_type.? != Value) {
				@compileError("`init` fn on " ++ value_name ++ " must return a " ++ value_name);
			}

			const concrete_argument = @as(*Value, @ptrCast(@alignCast(argument)));

			if (init_type.Fn.params.len != 1 or init_type.Fn.params[0].type.? != *Value.Param) {
				@compileError("`init` fn on " ++ value_name ++ " must accept a " ++ @typeName(*Value.Param));
			}

			concrete_argument.* = Value.init(&@as(*ValueParams.Node, @ptrCast(@alignCast(erased_node))).param);
		}

		fn unbind(allocator: std.mem.Allocator, erased_node: *anyopaque, context: World.UnbindContext) void {
			if (@hasDecl(Value, "unbind")) {
				const node = @as(*ValueParams.Node, @ptrCast(@alignCast(erased_node)));

				if (node.has_prev) |prev| {
					prev.has_next = node.has_next;
				}

				if (node.has_next) |next| {
					next.has_prev = node.has_prev;
				}

				if (context.world.get_state(ValueParams)) |params| {
					if (node.has_prev == null) {
						params.has_head = node.has_next;
					}

					if (node.has_next == null) {
						params.has_tail = node.has_prev;
					}
				}

				Value.unbind(&node.param, context);
				allocator.destroy(node);
			}
		}
	};

	return comptime &.{
		.thread_restriction = if (@hasDecl(Value, "thread_restriction")) Value.thread_restriction else .none,
		.init = parameters.init,
		.bind = parameters.bind,
		.unbind = parameters.unbind,
	};
}

pub fn start(comptime setup: Setup, comptime options: Options) fn () anyerror!void {
	const Start = struct {
		fn main() anyerror!void {
			defer {
				heap.trace_leaks();
			}

			parse_args: for (std.os.argv[1 ..]) |arg| {
				const arg_span = std.mem.span(arg);
				const arg_split_index = std.mem.indexOfScalar(u8, arg_span, '=') orelse arg_span.len;
				const arg_name = arg_span[0 .. arg_split_index];

				for (LaunchFlag.values) |value| {
					const name = @tagName(value);

					if (!std.mem.eql(u8, arg_name, name)) {
						continue;
					}

					LaunchFlag.args[@intFromEnum(value)] =
						if (arg_split_index == arg_span.len)
							name
						else
							arg_span[arg_split_index ..];

					continue: parse_args;
				}
			}

			var world = try switch (options.execution) {
				.single_threaded => World.init(0),

				.thread_share => |thread_share| init: {
					const cpu_count = @as(u32, @intCast(std.math.clamp(std.Thread.getCpuCount() catch |cpu_count_error| {
						@panic(switch (cpu_count_error) {
							error.PermissionDenied => "permission denied retrieving CPU count",
							error.SystemResources => "system resources are preventing retrieval of the CPU count",
							error.Unexpected => "unexpected failure retrieving CPU count",
						});
					}, 0, std.math.maxInt(u32))));

					break: init World.init(scalars.fractional(cpu_count, thread_share) orelse 0);
				},
			};

			defer world.deinit();

			const events = App.Events{
				.load = try world.create_event("load"),
				.pre_update = try world.create_event("pre-update"),
				.update = try world.create_event("update"),
				.post_update = try world.create_event("post-update"),
				.render = try world.create_event("render"),
				.finish = try world.create_event("finish"),
				.exit = try world.create_event("exit"),
			};

			const app = try world.set_get_state(App{
				.events = &events,
				.target_frame_time = 1.0 / @as(f64, @floatFromInt(options.tick_rate)),
				.elapsed_time = 0,
				.is_running = true,
			});

			inline for (options.middleware) |setup_middleware| {
				try setup_middleware(&world, events);
			}

			try setup(&world, events);
			try world.run_event(events.load);

			const ticks_initial = std.time.milliTimestamp();
			var ticks_previous = ticks_initial;
			var accumulated_time = @as(f64, 0);

			while (app.is_running) {
				const ticks_current = std.time.milliTimestamp();
				const milliseconds_per_second = 1000.0;
				const delta_time = @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;

				app.elapsed_time = @as(f64, @floatFromInt(ticks_current - ticks_initial)) / milliseconds_per_second;
				ticks_previous = ticks_current;
				accumulated_time += delta_time;

				try world.run_event(events.pre_update);

				while (accumulated_time >= app.target_frame_time) : (accumulated_time -= app.target_frame_time) {
					try world.run_event(events.update);
				}

				try world.run_event(events.post_update);
				try world.run_event(events.render);
				try world.run_event(events.finish);
			}

			try world.run_event(events.exit);
		}
	};

	return Start.main;
}

pub fn system_fn(comptime call: anytype) *const SystemInfo {
	const Call = @TypeOf(call);

	const system_info = comptime generate: {
		switch (@typeInfo(Call)) {
			.Fn => |call_fn| {
				if (call_fn.params.len > SystemInfo.max_parameters) {
					@compileError("number of parameters to `call` cannot be more than 16");
				}

				const systems = struct {
					fn run(parameters: []const *const SystemInfo.Parameter, data: *const [SystemInfo.max_parameters]*anyopaque) anyerror!void {
						var call_args = @as(std.meta.ArgsTuple(Call), undefined);

						inline for (parameters, &call_args, data[0 .. parameters.len]) |parameter, *call_arg, state| {
							parameter.init(call_arg, state);
						}

						switch (@typeInfo(call_fn.return_type.?)) {
							.Void => @call(.auto, call, call_args),
							.ErrorUnion => try @call(.auto, call, call_args),
							else => @compileError("number of parameters to `call` must return void or !void"),
						}
					}
				};

				var parameters = @as([SystemInfo.max_parameters]*const SystemInfo.Parameter, undefined);
				var system_thread_restriction = ThreadRestriction.none;

				for (0 .. call_fn.params.len) |index| {
					const CallParam = call_fn.params[index].type.?;
					const parameter = parameter_type(CallParam);

					if (parameter.thread_restriction != .none) {
						if (system_thread_restriction != .none and system_thread_restriction != parameter.thread_restriction) {
							@compileError("a system may not have conflicting thread restrictions");
						}

						system_thread_restriction = parameter.thread_restriction;
					}

					parameters[index] = parameter;
				}

				break: generate &.{
					.parameters = parameters,
					.parameter_count = call_fn.params.len,
					.execute = systems.run,
					.thread_restriction = system_thread_restriction,
				};
			},

			else => @compileError("parameter `call` must be a function"),
		}
	};

	return system_info;
}

pub fn type_id(comptime T: type) TypeID {
	const TypeHandle = struct {
		comptime {
			_ = T;
		}

		var byte: u8 = 0;
	};

	return @enumFromInt(@intFromPtr(&TypeHandle.byte));
}

pub fn thread_restriction(comptime State: type) ThreadRestriction {
	if (@hasDecl(State, "thread_restriction")) {
		return State.thread_restriction;
	}

	return .none;
}
